# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

"""
This module contains all touch points with the package :py:mod`luigi`. We had to adapt the method
`run_luigi`, because we need to handle the system exit by ourselves.

The current logic of luisy looks looks like this:

- Do some luisy stuff like hashing every Task in the new DAG built up by given root task
- Run luigi
- Do some luisy stuff, e.g. save hashes.

"""

import luigi
import sys
from luigi.cmdline_parser import CmdlineParser
from luigi.setup_logging import InterfaceLogging
from luigi.retcodes import retcode
import logging

from luisy.testing import debug_run
from luisy.config import Config
from luisy.hashes import (
    HashSynchronizer,
    get_upstream_tasks,

)


class COLORS:
    RED = '\033[91m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    END = '\033[0m'


logger = logging.getLogger('luigi-interface').getChild("luisy-interface")


def run_luigi(argv):
    """
    Run luigi with command line parsing. This function is taken from
    :py:func:`luigi.retcodes.run_with_retcodes()`.
    We need to overwrite this,because luigi calls sys.exit(). luisy should keep on running to do
    stuff after luigi completed.

    Args:
        argv: Should (conceptually) be ``sys.argv[1:]``

    """
    logger = logging.getLogger('luigi-interface')
    with luigi.cmdline_parser.CmdlineParser.global_instance(argv):
        retcodes = retcode()

    try:
        worker = luigi.interface._run(argv).worker
    except luigi.interface.PidLockAlreadyTakenExit:
        sys.exit(retcodes.already_running)
    except Exception:
        # Some errors occur before logging is set up, we set it up now
        env_params = luigi.interface.core()
        InterfaceLogging.setup(env_params)
        logger.exception("Uncaught exception in luigi")
        sys.exit(retcodes.unhandled_exception)

    with luigi.cmdline_parser.CmdlineParser.global_instance(argv):
        task_sets = luigi.execution_summary._summary_dict(worker)
        root_task = luigi.execution_summary._root_task(worker)
        non_empty_categories = {k: v for k, v in task_sets.items() if v}.keys()

    def has(status):
        assert status in luigi.execution_summary._ORDERED_STATUSES
        return status in non_empty_categories

    codes_and_conds = (
        (retcodes.missing_data, has('still_pending_ext')),
        (retcodes.task_failed, has('failed')),
        (retcodes.already_running, has('run_by_other_worker')),
        (retcodes.scheduling_error, has('scheduling_error')),
        (retcodes.not_run, has('not_run')),
    )
    expected_ret_code = max(code * (1 if cond else 0) for code, cond in codes_and_conds)

    failed_ret_code = expected_ret_code == 0
    root_task_not_complete = root_task not in task_sets["completed"]
    root_task_not_already_done = root_task not in task_sets["already_done"]
    is_failed = failed_ret_code and root_task_not_complete and root_task_not_already_done
    return is_failed, expected_ret_code, task_sets, root_task


class CmdPrinter:
    def __init__(self, hash_synchronizer):
        self.hash_synchronizer = hash_synchronizer

    def print_tasks(self, text, hash_entries=None):
        if hash_entries is None:
            hash_entries = []
        print(text)
        for hash_entry in hash_entries:
            size = len(hash_entry.task.__repr__())
            print(
                COLORS.BLUE, " ", hash_entry.task, COLORS.END,
                "-" * (30 - size) + "->",
                COLORS.GREEN, hash_entry.filename, COLORS.END)

        print("\n")

    def print_opening(self, message=None):
        print(COLORS.RED, "=> Hey you! <=", COLORS.END)
        if message is not None:
            print(COLORS.RED, message, COLORS.END)

    def print_hash_changes(self):
        self.print_tasks(
            text="The following hashes are out-dated on your system:",
            hash_entries=self.hash_synchronizer.get_outdated_local_hash_entries())

    def print_new_hashes(self):
        self.print_tasks(
            text='The following hashes do not exists on your system:',
            hash_entries=self.hash_synchronizer.get_new_local_hash_entries()
        )

    def print_sources(self):
        # Not downloadable but new or outdated
        self.print_tasks(
            text="The following tasks will be executed:",
            hash_entries=self.hash_synchronizer.get_local_execution_hash_entries())
        if Config().get_param('download'):
            self.print_tasks(
                text="The following files can be downloaded from the cloud (only necessary ones "
                     "will be):",
                hash_entries=self.hash_synchronizer.get_download_hash_entries())

    def print_uploads(self):
        if Config().get_param('upload'):
            self.print_tasks(
                text="The following files will be uploaded to the cloud and existing files "
                     "overwritten",
                hash_entries=self.hash_synchronizer.get_upload_hash_entries()
            )

    def print_cloud_hash_changes(self):
        self.print_tasks(
            text="The following hashes are outdated in the cloud and will be overwritten:",
            hash_entries=self.hash_synchronizer.get_outdated_cloud_hash_entries()
        )

    def get_user_permission(self, message=None):
        """
        Asks the user if the change applied is alright. These checks can be skipped if the
        parameter `no_ask` is set.
        """
        if not Config().get_param('no_ask'):
            user_input = input("Type 'y' to proceed: ")
            if user_input.lower() != 'y':
                print(f"User input '{user_input}' != y -> Bye")
                sys.exit(0)
        else:
            logger.info('--no-ask set, wont ask anything')


class Coordinator(object):
    """

    Parent class to handle different luisy modes. The basic coordinator does hash computation of
    a given root task and reads hashes that are found on the local filesystem in `.luisy.hash`
    file.

    Args:
        args_luisy (dict): Arguments for luisy
        root_task (luisy.tasks.base.Task): Root task to be executed

    """

    def __init__(self, args_luisy, root_task=None):
        self.args_luisy = args_luisy
        self.root_task = root_task

        Config().update(self.args_luisy)
        Config().check_params()

        # Get all tasks
        self.tasks = get_upstream_tasks(self.root_task)

    def initialize_hash_synchronization(self):
        self.hash_synchronizer = HashSynchronizer(
            tasks=self.tasks,
            root_task=self.root_task
        )
        self.hash_synchronizer.initialize()

    def initialize_printer(self):
        self.printer = CmdPrinter(self.hash_synchronizer)

    def run(self):
        self.initialize_hash_synchronization()
        self.initialize_printer()
        self.print_approvals()
        self.hash_synchronizer.set_files_to_download()

        if Config().get_param('dry_run'):
            print('This is just a dry-run. Byebye!')
            return 0

        run_failed = self.run_coordinator()

        if run_failed:
            return 1
        else:
            return 0

    def run_coordinator(self):
        raise NotImplementedError()

    def print_approvals(self):
        raise NotImplementedError()


class HashUpdateMode(Coordinator):
    """
    The HashUpdateMode is started when the user passes `--hash-update` to luisy. Instead of
    rerunning all the tasks, whose hash changed in this run, this mode only updates the hashes (
    which already exist) in the `.luisy.hash` file and does not run the tasks again. This feature
    can be useful when refactoring your luisy Pipeline. It's also possible to update the hashes
    in the cloud when `--upload` is passed to luisy.
    """

    def print_approvals(self):
        self.printer.print_opening(
            message='You are running in Hash-Update-Mode! This means no '
                    'tasks will be run, downloaded or uploaded. Only the '
                    'hashes in luisy.hash files will be updated locally '
                    'and in the cloud if --download flag is set'
        )
        self.printer.print_hash_changes()
        self.printer.print_cloud_hash_changes()
        self.printer.get_user_permission()

    def run_coordinator(self):
        self.hash_synchronizer.synchronize_local_hashes(failed_tasks=[])
        if Config().get_param('upload'):
            self.hash_synchronizer.synchronize_cloud_hashes(failed_tasks=[], upload_tasks=False)


class PipelineMode(Coordinator):
    """
    This class manages the communication between luisy and luigi functionalities.

    Args:
        args_luisy (dict): Arguments for luisy
        args_luigi (list): Parsed command line arguments for luigi
        root_task (luisy.tasks.base.Task): Root task to be executed
        run_luigi_cli (bool): If true, luigi is called via command line, otherwise
            :py:func:`luisy.testing.debug_run` is called.

    """

    def __init__(self, args_luisy, args_luigi, root_task=None, run_luigi_cli=True):
        self.args_luigi = args_luigi
        self.run_luigi_cli = run_luigi_cli

        super().__init__(args_luisy=args_luisy, root_task=root_task)

    def run_luigi(self):

        if self.run_luigi_cli:
            luigi_failed, _, self.tasks_set, _ = run_luigi(self.args_luigi)
            return luigi_failed
        else:
            print('RUNNING')
            print(self.root_task)
            self.tasks_set = debug_run([self.root_task], return_summary=True)
            return len(self.tasks_set['failed']) > 0

    def finalize_luigi_run(self):
        self.hash_synchronizer.synchronize_local_hashes(failed_tasks=self.tasks_set['failed'])
        if Config().get_param('upload'):
            self.hash_synchronizer.synchronize_cloud_hashes(failed_tasks=self.tasks_set['failed'])

    def run_coordinator(self):
        self.hash_synchronizer.clear_outdated_tasks()
        luigi_failed = self.run_luigi()
        self.finalize_luigi_run()
        return luigi_failed

    def print_approvals(self):
        self.printer.print_opening('You are in Pipeline Mode!')
        self.printer.print_hash_changes()
        self.printer.print_new_hashes()
        self.printer.print_sources()
        self.printer.print_uploads()
        self.printer.get_user_permission()


def get_root_task(args_luigi):
    with CmdlineParser.global_instance(args_luigi) as cp:
        return cp.get_task_obj()



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0
"""
This module contains the entrypoint of luisy. Here we catch the CLI arguments of the user and
parse them inside a function. We also separate luigi params from luisy params here. After setting up
the luisy parameters in our :py:class:`luisy.config.Config` singleton, the logic of luisy is
called.
"""

import argparse
import logging
import sys
from luisy.luigi_interface import (
    PipelineMode,
    HashUpdateMode,
    get_root_task,
)
from luisy.config import get_default_params

logger = logging.getLogger(__name__)


def disable_azure_logging():
    log = logging.getLogger("azure.core.pipeline.policies.http_logging_policy")
    log.setLevel(logging.ERROR)


def parse_args(argv):
    """
    Method to get all luisy args out of the argv given inside the cli. Unknown arguments will be
    returned separately, to later forward them to luigi

    Args:
        argv: Should (conceptually) be ``sys.argv[1:]``

    Returns:
        tuple: dict of parsed luisy arguments and rest of
            `argv` in an unparsed format

    """
    parser = argparse.ArgumentParser(description='luisy Parser')
    for key, val in get_default_params().items():
        key = key.replace('_', '-')
        if type(val) == bool:
            parser.add_argument('--' + key, default=val, action='store_true')
        else:
            parser.add_argument('--' + key, default=val)

    args_luisy, args_luigi = parser.parse_known_args(argv)
    args_luisy = {key: val for key, val in vars(args_luisy).items() if val is not None}
    return args_luisy, args_luigi


def luisy_run(argv=sys.argv[1:]):
    """
    luisy CLI Call

    Args:
        argv: Should (conceptually) be ``sys.argv[1:]``

    """

    disable_azure_logging()
    args_luisy, args_luigi = parse_args(argv)
    root_task = get_root_task(args_luigi)
    force_hash_update = args_luisy.pop('hash_update_mode')
    if force_hash_update:
        ret_code = HashUpdateMode(
            args_luisy=args_luisy,
            root_task=root_task
        ).run()
    else:
        # Run the pipeline
        ret_code = PipelineMode(
            args_luisy=args_luisy,
            args_luigi=args_luigi,
            root_task=root_task,
        ).run()

    sys.exit(ret_code)


def build(task, workers=1, local_scheduler=True, **args_luisy):
    """
    Similar to :py:func:`luigi.build`, it executes a task and along with upstream requirements.
    There, it uses the hashing functionality to detect tasks that need a rerun.

    Args:
        task (luisy.base.Task): The luisy task to be executed (along with its upstream
            requirements).
        workers (int): Number of (luigi) workers
        local_scheduler (bool): Whether to use the local scheduler of luigi

    Returns:
        luisy.luigi_interface: Coordinator: The coordinator object of the luisy/luigi build
    """

    args_luigi = [f"--workers={workers}"]

    if local_scheduler:
        args_luigi.append('--local-scheduler')

    if 'no_ask' not in args_luisy:
        args_luisy['no_ask'] = True

    if 'hash_update_mode' not in args_luisy:
        hash_update_mode = False
    else:
        hash_update_mode = args_luisy['hash_update_mode']

    if hash_update_mode:
        coordinator = HashUpdateMode(
            args_luisy=args_luisy,
            root_task=task
        )
    else:
        # Run the pipeline
        coordinator = PipelineMode(
            args_luisy=args_luisy,
            args_luigi=args_luigi,
            root_task=task,
            run_luigi_cli=False
        )

    coordinator.run()
    return coordinator



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0
"""
This module holds various decorators to be attached to :py:class:`~luisy.tasks.base.Task` objects.
"""
import os
from luisy.config import Config
from luisy.targets import (
    HDFTarget,
    PickleTarget,
    XLSXTarget,
    CSVTarget,
    ParquetDirTarget,
    JSONTarget,
    DirectoryTarget,
)
from luigi.util import inherits
from luigi import ExternalTask
from luisy.helpers import get_string_repr


class requires(object):
    """
    A backwards compatible extension to :py:class:`luigi.util.requires` which allows to access the
    requirements in a dict like fashion.
    """

    def __init__(self, *tasks_to_require, as_dict=False):
        if not tasks_to_require:
            raise TypeError("tasks_to_require cannot be empty")
        self.tasks_to_require = tasks_to_require
        self.as_dict = as_dict

    def __call__(self, task_that_requires):
        task_that_requires = inherits(*self.tasks_to_require)(task_that_requires)

        def _requires(_self):
            """
            """
            if len(self.tasks_to_require) == 1:
                return _self.clone_parent()
            else:
                if self.as_dict:

                    return {
                        t.get_task_family(): _self.clone(t) for t in self.tasks_to_require
                    }

                else:
                    return _self.clone_parents()

        task_that_requires.requires = _requires
        return task_that_requires


def _make_output_decorator_factory(target_cls):
    def output_decorator_factory(cls_py=None, **kwargs):
        def output_decorator(cls):
            cls.target_cls = target_cls
            cls.target_kwargs = kwargs
            return cls
        return output_decorator(cls_py) if callable(cls_py) else output_decorator
    return output_decorator_factory


hdf_output = _make_output_decorator_factory(HDFTarget)
pickle_output = _make_output_decorator_factory(PickleTarget)
xlsx_output = _make_output_decorator_factory(XLSXTarget)
csv_output = _make_output_decorator_factory(CSVTarget)
json_output = _make_output_decorator_factory(JSONTarget)


def parquetdir_output(cls_py=None):
    def decorator(cls):
        """
        """
        cls.target_cls = ParquetDirTarget

        if not issubclass(cls, ExternalTask):
            raise ValueError(
                "This decorator only works on ExternalTasks"
            )

        if not hasattr(cls, 'get_folder_name'):
            raise ValueError(
                "When using ParquetDirTarget, a method get_folder_name()"
                "needs to be implemented in your luisy task"
            )

        def get_outfile(self):
            return os.path.join(
                self.get_outdir(),
                self.get_folder_name()
            )

        cls.get_outfile = get_outfile
        return cls

    return decorator(cls_py) if callable(cls_py) else decorator


def make_directory_output(read_file_func, regex_pattern=None):
    """
    This function is used to give the user the ability to create his own directory outputs for
    external Tasks. Decorating a :py:class:ExternalTask with this function allows the user to use
    :py:meth:`luisy.ExternalTask.output().read` in the task that requires the external task. The
    files will be passed one by one as a generator when calling task.input().read().

    Args:
        read_file_func: function read_file(file), where the user can define how his data is being
            loaded. return the filecontent in this function. To skip specific incoming files, make
            sure that your `read_file_func` function returns None.
        regex_pattern (str): regex string that can be passed. All files found in the directory
          will be checked against this regex and only be parsed if it matches.

    Returns:
        func: Decorator that can be attached to :py:class:`ExternalTask`
    """

    def directory_output(cls):
        if not issubclass(cls, ExternalTask):
            raise ValueError(
                "This decorator only works on ExternalTasks"
            )
        if not hasattr(cls, 'get_folder_name'):
            raise ValueError(
                "When using the directory_output decorator, a method get_folder_name()"
                "needs to be implemented in your luisy task"
            )

        cls.target_cls = DirectoryTarget

        def read_file(self, file):
            return read_file_func(file)

        def get_outfile(self):
            return os.path.join(
                self.get_outdir(),
                self.get_folder_name()
            )

        cls.target_cls.read_file = read_file
        cls.target_cls.regex_pattern = regex_pattern
        cls.get_outfile = get_outfile
        return cls

    return directory_output


def project_name(project_name):
    """
    Sets the project_name of the task to `project_name` to overwrite the default behavior of
    :py:class:`luisy.Task`

    Args:
        project_name (str): Name of the project
    """

    def decorator(cls):
        cls.project_name = project_name
        return cls

    return decorator


def _make_dir_layer(layer_name):
    def dir_layer(cls_py=None):
        def decorator(cls):
            def get_outdir(self):
                out_dir = os.path.join(
                    Config().get_param('working_dir'), self.project_name, layer_name)
                out_dir = os.path.join(
                    out_dir, self.get_sub_dir())
                return out_dir

            cls.get_outdir = get_outdir
            return cls
        return decorator(cls_py) if callable(cls_py) else decorator
    return dir_layer


def auto_filename(cls_py=None, excl_params=None, excl_none=False):
    """
    "Factory" for a Decorator which automatically generates the basename of the outfile based on
    the name of the class and the :py:class:`luigi.Parameter` objects attached to it. Parameters
    allow the user to exclude some parameters and also to exclude None values in general

    Args:
        cls_py (class): class that gets decorated. This parameter is kind of a hack to allow
            usage of this decorator without parenthesis "()". So the end-user is able to call
            `@auto_filename` (important for backwards compatibility),
            `@auto_filename(excl_list=['x', 'y'])` and `@auto_filename()` as well with all having
            the class below decorated
        excl_params (list): List of task parameters that should be excluded
        excl_none (bool): If True, None values are not part of the filename

    Returns:
        class: Decorated class

    """
    assert callable(cls_py) or cls_py is None
    if excl_params is None:
        excl_params = []

    def _decorator(cls):
        def get_file_name(self):
            filename = self.get_task_family()
            for p in sorted(self.get_param_names()):
                if p not in excl_params and p != 'working_dir':
                    if excl_none and getattr(self, p) is None:
                        continue
                    param_value = get_string_repr(getattr(self, p))
                    filename += f'_{p}={param_value}'
            return filename

        cls.get_file_name = get_file_name
        return cls

    return _decorator(cls_py) if callable(cls_py) else _decorator


interim = _make_dir_layer('interim')
final = _make_dir_layer('final')
raw = _make_dir_layer('raw')



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

default_params = {
    'download': False,
    'upload': False,
    'dry_run': False,
    'working_dir': None,
    'azure_storage_key': None,
    'azure_account_name': None,
    'azure_container_name': None,
    'no_ask': False,
    'requirements_path': None,
    'hash_update_mode': False,
}
env_keys = {
    'working_dir': 'WORKING_DIR',
    'requirements_path': 'LUISY_REQUIREMENTS_PATH',
    'azure_storage_key': 'LUISY_AZURE_STORAGE_KEY',
    'azure_account_name': 'LUISY_AZURE_ACCOUNT_NAME',
    'azure_container_name': 'LUISY_AZURE_CONTAINER_NAME',
}



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import networkx as nx
import numpy as np
import matplotlib.pyplot as plt

from functools import partial
from luigi.task import flatten


def simple_task_string(task):
    """
    Create simple string label out of task.
    """
    groups = task.task_id.split("_")
    return f"{groups[0]}\n{groups[-1]}"


def create_label(task, excl_list=(), with_id=False):
    """
    Make a string label out of a luigi task. Make sure the string is unique !
    Args:
        task (luisy.tasks.base.Task): The task the label should be created for
        excl_list (list): List of parameter kwargs that are not included in the label
        with_id (bool): If False, do not include the hash of the task.

    Returns:
        str: Label to be used for the node
    """
    groups = task.task_id.split("_")
    name = groups[0]
    if with_id:
        name += f"{groups[-1]}"
    for k in task.param_kwargs:
        if k not in excl_list:
            name += f"\n{k}={task.param_kwargs[k]}"
    return name


def add_children(task, graph, namefunc=create_label, depth=1, depth_limit=10,
                 unique_children=False):
    """
    Recursive Function that adds the children of task to a graph object. Make sure that 'namefunc'
    returns unique strings for all tasks that will be included in the tree
    Args:
        task (luisy.tasks.base.Task): The task
        graph (networkx.DiGraph): The directed ayclic graph
        namefunc (callable): Function that creates a string out a task.
        depth (int): Depth of current node
        depth_limit (int): Max allowed depth of tree
        unique_children (bool): If true, add only subset of unique types of children for each task
    """
    if depth <= depth_limit:
        children = flatten(task.requires())
        if unique_children:
            children = unique_types(children)
        current_depth = depth + 1
        for c in children:  # depth first search
            graph.add_edge(namefunc(task), namefunc(c))
            add_children(
                task=c,
                graph=graph,
                namefunc=namefunc,
                depth=current_depth,
                unique_children=unique_children,
                depth_limit=depth_limit,
            )


def make_dependency_graph(task, excl_list=(), namefunc=None, unique_children=True, depth_limit=10):
    """
    Create pydot graph out of a tasks dependencies.
     Args:
        task (luisy.tasks.base.Task): task
        excl_list (list): Parameter forwarded to 'create_label' if namefunc is none
        namefunc (callable): Function that creates a string out a task.
        depth_limit (int): Max allowed depth of tree
        unique_children (bool): If true, add only subset of unique types of children
    """
    dag = nx.DiGraph()
    if namefunc is None:
        namefunc = partial(create_label, excl_list=excl_list)
    dag.add_node(namefunc(task))  # if task has no children the root node has to be created
    add_children(task, dag, namefunc, unique_children=unique_children, depth_limit=depth_limit)
    return dag


def unique_types(tasks):
    """
    Given a list of objects, return a subset that has each type only once.
    """
    types = []
    res_list = []
    for e in tasks:
        name = type(e).__name__
        if name not in types:
            types.append(name)
            res_list.append(e)
    return res_list


def visualize_task(task, ax=None, unique_children=True, parameters_to_exclude=()):
    """
    Visualizes the dependencies of a :py:class:`luisy.Task`.

    Args:
        task (luisy.tasks.base.Task): The task whose dependencies should be visualized.
        ax (matplotlib.axis.Axis): A matplotlib axis to draw in
        unique_children (bool): If true, add only subset of unique types of children
        parameters_to_exclude (list): List of names of py:mod:`luigi`-parameters to exclude in
            visualization.
    """

    dag = make_dependency_graph(
        task=task,
        unique_children=unique_children,
        depth_limit=np.inf,
        excl_list=parameters_to_exclude,
    )

    if ax is None:
        _, ax = plt.subplots(figsize=(15, 10))

    pos = nx.shell_layout(dag)

    nx.draw_networkx_labels(dag, ax=ax, pos=pos)
    nx.draw_networkx_edges(dag, ax=ax, pos=pos)
    ax.axis('off')



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import ast
import json
import hashlib
import inspect
import importlib
import logging
import os
import subprocess
import sys

from distlib.database import DistributionPath
import numpy as np
import pandas as pd
import requirements

logger = logging.getLogger(__name__)


class RequirementFileNotFound(Exception):
    pass


def create_hashes(task_list, requirements_path=None):
    """
    Produces a hash for each luisy Task in the 'task_list'.
    The hash changes as soon as any of the code that is inside the ClassDef of a task (or any
    other code used by the task) changes.

    Args:
        task_list (list[luisy.Task]): List of task classes
        requirements_path (str): location if requirements.txt, if None, try to get it automatically

    Returns:
        list[str]: The hash values for the tasks in hexadecimal.
    """
    hashes_list = list()

    if requirements_path is None:
        #  This should work with any task from the list
        task_class = task_list[0]
        try:
            requirements_path = get_requirements_path(task_class)
            requirements_dict = get_requirements_dict(requirements_path)
        except RequirementFileNotFound:
            logger.warning(
                'No requirements.txt found.'
                'Will try to infer from installed version'
            )
            requirements_dict = get_all_deps_with_versions(
                task_class.__module__.split('.')[0],
            )
    else:
        requirements_dict = get_requirements_dict(requirements_path)

    deps_map = create_deps_map(requirements_dict)  # req -> dep, one line per dep

    for task_class in task_list:
        module = inspect.getmodule(task_class)
        hashes, imported_packages, vis_data = walk_nodes(task_class.__name__, module.__name__)
        df_reqs = map_imports_to_requirements(imported_packages, requirements_dict, deps_map)

        # Exclude luisy, as we don't want that a bugfix or new feature in luisy leads to
        # re-execution of all pipelines
        df_reqs = df_reqs[df_reqs['package'] != 'luisy']

        version_info_to_hash = df_reqs.requirement_line.unique().tolist()
        version_info_to_hash.sort()  # order comes from a set
        to_hash = "".join(hashes + version_info_to_hash)
        final_hash = hashlib.md5(to_hash.encode()).hexdigest()
        hashes_list.append(final_hash)

    return hashes_list


def walk_nodes(variable_name, module_name, parent="", parent_module=""):
    """
    Function that recursively visits AST nodes.
    Theses nodes are either ClassDefs, FunctionDefs or simply Name nodes.
    For each node

    * the sourcecode within the node body is hashed and added to a return list
    * the names of external packages used inside the node body is added to a return list
    * if there are variable names coming from outside the node body, but from the same package as
      the node, the function also visits these nodes

     Args:
        module_name (str): module name of current node
        parent (str): name of parent node
        parent_module (str): name of module of parent node
        variable_name (str): name of AST node. Class name, Function name, or a module variable.

    Returns:
        tuple: Triple of list of hashes, set of strings representing the imported packages, and a
        list of dicts holding the visualization data collected during traversal of
        the ast syntax tree. Each dict represents an AST node. Contains names of node and its
        parent, modules names of node and parent and node type. This info can be used to plot
        the dependency graph.
    """
    visualization_data = []

    module = importlib.import_module(module_name)
    if not hasattr(module, variable_name):
        raise AttributeError(f"Module '{module_name}' does not contain any variable named "
                             f"'{variable_name}'!")
    attr_type = type(getattr(module, variable_name)).__name__
    if attr_type == 'module':  # check if variable is itself a module
        module_name = f"{module_name}.{variable_name}"
        module = importlib.import_module(module_name)
    module_source = inspect.getsource(module)
    module_tree = ast.parse(module_source)
    node = module_tree  # node is the whole module

    if attr_type != 'module':
        node = get_node(variable_name, module_tree)  # node is a branch of the module tree

    visualization_data.append(
        {
            "name": variable_name,
            "parent": parent,
            "module": module_name,
            "parent_module": parent_module, "type": type(node).__name__
        }
    )
    hashes = [produce_node_hash(node)]

    # df containing all imports in module
    df_imports = get_import_data(module_tree)
    # all variables used inside node, but defined outside
    var_names = get_varnames_delta(node)

    #  filter imports that are used inside node
    df_relevant_imports = df_imports.loc[
        np.isin(df_imports.used_name, var_names)]  # .loc keeps columns in empty case
    df_relevant_imports = df_relevant_imports.loc[~ df_relevant_imports.package.apply(is_standard)]
    # join external imports with requirements file
    package = module_name.split(".")[0]
    df_internal_imports = df_relevant_imports.loc[df_relevant_imports.package.values == package]
    imported_packages = set(df_relevant_imports.package)
    imported_packages.discard(package)  # imports of external libraries

    # filter variables that come from the module itself
    vars_same_module = var_names[~np.isin(var_names, df_imports.used_name)]

    # call function recursively on package-internal nodes
    varnames_next_level = vars_same_module.tolist() + df_internal_imports.varname.tolist()
    module_names_next_level = [module_name] * len(
        vars_same_module) + df_internal_imports.module.tolist()

    for var_, mod_ in zip(varnames_next_level, module_names_next_level):
        # drop special case like __name__  , or recursive calls
        if not var_.startswith("__") and var_ != variable_name:
            hashes_c, imports_c, digraph_data_c = walk_nodes(var_, mod_, parent=variable_name,
                                                             parent_module=module_name)
            hashes = hashes + hashes_c
            imported_packages.update(imports_c)
            visualization_data = visualization_data + digraph_data_c
    return hashes, imported_packages, visualization_data


def _remove_docstrings(node):
    """
    Removes the doc-strings
    """

    if isinstance(node, (ast.FunctionDef, ast.ClassDef, ast.Module)):
        if node.body and isinstance(node.body[0], ast.Expr):
            if isinstance(node.body[0].value, ast.Str):
                # Cut-off that element from the body
                # The reason why we do not set it to the empty string is that we also want to have
                # the same doc-string if someone adds a doc-string, not just changes it
                node.body = node.body[1:]

    if hasattr(node, 'body'):
        for _node in node.body:
            _remove_docstrings(_node)


def produce_node_hash(class_node):
    """
    Produce hash for AST node, normally ClassDef or FunctionDef or Constant.
    invariant to: comments, spaces, blank lines.
    not invariant to: variable renaming
    """

    _remove_docstrings(class_node)
    st = ast.dump(class_node, annotate_fields=False)  # no need to include unambiguous field names.
    hash_ = hashlib.md5(st.encode())
    return hash_.hexdigest()


def get_import_data(module_tree):
    """
    Get imports of a module.
    Args:
        module_tree(ast.AST): Syntax tree of module
    Returns:
        pandas.DataFrame: Data about imports to the module.
    """
    v = ImportReturner()
    v.visit(module_tree)
    data = v.import_data
    if len(data) > 0:
        df = pd.DataFrame(data)
        df["used_name"] = df.apply(lambda row: row.varname if row.asname is None else row.asname,
                                   axis=1)
        df["package"] = df.module.apply(lambda st: st.split(".")[0])
    else:
        df = pd.DataFrame(columns=["module", "varname", "asname", "used_name", "package"])

    return df


class ImportReturner(ast.NodeVisitor):
    """
    Scans syntax tree for imports.
    """

    def __init__(self):
        self.import_data = []

    def star_import_in_imports(self, lst):
        return any([import_node["varname"] == '*' for import_node in lst])

    def generic_visit(self, node):
        imports = self.filter_imports(node)
        if imports is not None:
            if self.star_import_in_imports(imports):
                raise Exception("Star import not supported by luisy hash creation!")
            self.import_data = self.import_data + imports
        ast.NodeVisitor.generic_visit(self, node)

    def filter_imports(self, node):
        if type(node).__name__ == "Import":
            aliases = node.names
            return [{"module": alias.name, "varname": alias.name, "asname": alias.asname} for alias
                    in
                    aliases]
        if type(node).__name__ == "ImportFrom":
            mod = node.module
            aliases = node.names
            return [{"module": mod, "varname": alias.name, "asname": alias.asname} for alias in
                    aliases]


def get_varnames_delta(node):
    """
    Find all variable names that are used the node body, but defined outside the node body.
    """
    vr = VariablesReturner()
    vr.visit(node)
    df_names = pd.DataFrame(vr.variable_names)  # lists loaded and stored varnames inside node
    if len(df_names) == 0:
        return np.array([])
    df_load = df_names.query("kind == 'Load'")
    df_store = df_names.loc[np.isin(df_names["kind"], ["Store", "arg", "FunctionDef", "ClassDef",
                                                       "ExceptHandler"])]
    df_load = df_load.groupby("id").agg({"lineno": "min"}).reset_index()
    df_store = df_store.groupby("id").agg({"lineno": "min"}).reset_index()
    df_match = df_store.merge(df_load, on="id")

    ids_to_ignore = list(df_match.id) + ["self"]
    df_names_outside_classdef = df_load.loc[~np.isin(df_load.id, ids_to_ignore)]
    filter_builtins = df_names_outside_classdef.id.apply(is_builtin)
    df_names_outside_classdef = df_names_outside_classdef.loc[~filter_builtins]

    return df_names_outside_classdef.id.values


class VariablesReturner(ast.NodeVisitor):
    """
    Scans syntax tree for variable names. While visiting all nodes of the ast syntax tree,
    save all names that are loaded or stored. These include FunctionDefs, ClassDefs, Function
    arguments or the Exception Handler context name.
    """

    def __init__(self):
        self.variable_names = []

    def filter_variables(self, node):
        if type(node).__name__ == "Name":
            ctx = node.ctx
            return {"id": node.id, "kind": type(ctx).__name__, "lineno": node.lineno}
        if type(node).__name__ == "arg":
            return {"id": node.arg, "kind": "arg", "lineno": node.lineno}

        #  Function Defs and Class Defs can also be inside another class def
        if type(node).__name__ == "FunctionDef":
            return {"id": node.name, "kind": "FunctionDef", "lineno": node.lineno}
        if type(node).__name__ == "ClassDef":
            return {"id": node.name, "kind": "ClassDef", "lineno": node.lineno}
        if type(node).__name__ == "ExceptHandler":
            return {"id": node.name, "kind": "ExceptHandler", "lineno": node.lineno}

    def generic_visit(self, node):
        var_names = self.filter_variables(node)
        if var_names is not None:
            self.variable_names.append(var_names)
        ast.NodeVisitor.generic_visit(self, node)


def get_node(node_name, module_tree):
    """
    Find an node with given name in the upper level of a module.
    """
    for node in module_tree.body:
        if type(node).__name__ == "Assign":
            target_names = [t.id for t in node.targets]
            if node_name in target_names:
                return node.value
        if type(node).__name__ in ["FunctionDef", "ClassDef"]:
            if node.name == node_name:
                return node
    raise ValueError(f"Node '{node_name}' was not found in AST tree.")


def get_requirements_path(task_class):
    """
    Tries to automatically find the path of requirements file.
    """
    path_ = inspect.getfile(task_class)
    while True:
        groups = path_.split(os.sep)
        if len(groups) == 1:
            raise RequirementFileNotFound(
                "Requirements file not could not be found automatically!"
            )
        path_ = os.sep.join(groups[:-1])
        req_path = os.path.join(path_, "requirements.txt")
        if os.path.exists(req_path):
            break
    return req_path


def get_requirements_dict(path):
    """
    Open the requirements file and create a dictionary which maps each package to its
    corresponding line in the requirements.txt. This line contains all version info and will be used
    for the hash later.
    """
    reqs_dict = {}
    with open(path, "r") as f:
        for req in requirements.parse(f):
            reqs_dict[req.name] = req.line.strip()
    return reqs_dict


def _harmonize_package_name(package_name):
    return package_name.replace("-", "_")


def create_deps_map(requirements_dict):
    """
    Map each requirement to its dependencies.
    Both keys and values are in unified format. That means to apply: .lower().replace("-","_")
    """
    deps_map = {}
    for req in requirements_dict:  # here we need the Pypi name
        deps = [_harmonize_package_name(d.lower()) for d in get_all_deps(req)]
        deps_map[_harmonize_package_name(req.lower())] = deps
    return deps_map


def get_all_deps_with_versions(package):
    """
    Get names of all dependencies of 'package' using 'pipdeptree' including their required version.

    Args:
        package(str): package for which we want deps.

    Returns:
        dict: Dict where the keys are the package names and the values are the version constraints
        of all dependencies.
    """

    # Underlines in package names become hyphens due to a strange behavior of setuptools
    package = package.replace('_', '-')
    reqs = subprocess.check_output(
        [
            sys.executable,
            "-m", "pipdeptree", "--package",
            package,
            "-w", "silence", "--json-tree"
        ],
        stderr=subprocess.DEVNULL,
    )
    reqs_list = json.loads(reqs)

    if len(reqs_list) == 0:
        return {}

    reqs_list = reqs_list[0].get('dependencies', [])

    reqs_dict = {
        _harmonize_package_name(req['package_name']): "".join(
            [
                _harmonize_package_name(req['package_name']),
                req['required_version'].replace('Any', ''),
            ]) for req in reqs_list
    }

    return reqs_dict


def get_all_deps(package):
    reqs_dict = get_all_deps_with_versions(package)
    return set(reqs_dict.keys())


def map_imports_to_requirements(imported_packages, requirements_dict, deps_map):
    """
    Given a list of packages, look whether they are in the requirements of the repository.
    If it is not a direct requirement, look if it is a dependency of a requirements.
    Args:
        imported_packages(iterable[str]): list of package names
        requirements_dict(dict): dict returned by ...
        deps_map(dict): dict returned by

    Returns:
        pandas.DataFrame: Dictionary listing for each package, one or several lines of the
        requirements file that concern it.
    """
    package_name_to_pypi = make_pypi_mapper()
    imported_packages = [package_name_to_pypi(p) for p in imported_packages]
    requirements_dict = {k.lower().replace("-", "_"): v for (k, v) in requirements_dict.items()}
    imports_to_req = []
    for package in imported_packages:
        if package in requirements_dict:
            imports_to_req.append(
                {"package": package, "requirement_line": requirements_dict[package]})
        else:  # Not directly in requirements file.
            for req in deps_map:
                if package in deps_map[req]:
                    imports_to_req.append(
                        {"package": package, "requirement_line": requirements_dict[req]})

    if len(imports_to_req) == 0:
        df_reqs = pd.DataFrame(data={'package': [], 'requirement_line': []})
    else:
        df_reqs = pd.DataFrame(imports_to_req)
    is_in_reqs = np.isin(imported_packages, df_reqs.package)
    if not is_in_reqs.all():
        not_found = np.array(imported_packages)[~is_in_reqs]
        raise Exception(f"Imported packages {not_found} "
                        f"could not be mapped to requirements file")
    return df_reqs


def is_builtin(func):
    """
    This function is work in progress and will be improved with every upcoming error.
    """
    try:
        f = eval(func)
        if hasattr(f, "__module__"):
            if f.__module__ == 'builtins' or type(f).__name__ == 'builtin_function_or_method':
                return True
    except NameError:  # variable is not defined
        return False
    return False


def is_standard(package_name):
    """
    Determines whether a package is in the python standard lib.
    """
    if package_name in ["sys"]:  # did not work like for the others
        return True
    try:
        major = sys.version_info.major
        minor = sys.version_info.minor
        std_folder = f"python{major}.{minor}"
        case_1 = os.path.join(std_folder, f"{package_name}.py")
        case_2 = os.path.join(std_folder, package_name, "__init__.py")
        path = importlib.util.find_spec(package_name).origin
        if path is None:
            # have to actually load module to determine file.
            mod = importlib.__import__(package_name)
            path = mod.__file__
        if path.endswith(case_1) or path.endswith(case_2) or path == 'built-in':
            return True

    except Exception:
        return False

    return False


def make_pypi_mapper():
    special_pypi = get_irregular_pypi_names()

    def package_name_to_pypi(package):
        """
        If the standardized package name is not the Pypi name, return standardized pypi-name
        the node, the function also visits these nodes

         Args:
            package(str): name of python package, (standardized with .lower().replace("-", "_"))
        Returns:
            package(str): Pypi name (standardized with .lower().replace("-", "_")) if the latter
            differs from package name.
        """
        if package in special_pypi:
            return special_pypi[package]
        else:
            return package

    return package_name_to_pypi


def get_irregular_pypi_names():
    """
    Get pypi name for given package if they differ.
     unify everything to format lower().replace("-","_")  is not the package name
    """
    distlib_logger = logging.getLogger("distlib")
    distlib_logger.setLevel(40)  # too much info
    dist_path = DistributionPath(include_egg=True)
    map_ = {}
    for dist in dist_path.get_distributions():
        modules = [m.lower().replace("-", "_") for m in dist.modules]
        name = dist.name.lower().replace("-", "_")
        if len(modules) > 0 and modules != [name]:
            for m in modules:
                map_[m] = name
    return map_
